/* Copyright 2013 Thomas Prokosch

This file is part of pnmfix.

pnmfix is free software: you can redistribute it and/or modify it
under the terms of the GNU General Public License as published by the
Free Software Foundation, either version 3 of the License, or (at your
option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program. If not, see <http://www.gnu.org/licenses/>.
*/

package main

/* netpbm file format:
-  http://netpbm.sourceforge.net/doc/pbm.html
-  http://netpbm.sourceforge.net/doc/pgm.html
-  http://netpbm.sourceforge.net/doc/ppm.html
*/

import (
	"flag"
	"fmt"
	"io"
	"os"
	"strconv"
	"syscall"
	"unsafe"
)

// Convenience values to ease the reading of the imageHeader struct.
const (
	Magic = iota
	Width
	Height
	Maxval
)

// imageHeader contains the PNM image header as well as some
// computated values specific to the given image. The values taken
// directly from the PNM file are Magic, Width, Height, and Maxval.
// Resulting from Maxval is Depth; HeaderSize specifies the offset of
// the raw image data from the beginning of the file, while FileSize
// gives the total number of bytes in the file.
type imageHeader struct {
	Magic string
	Width, Height, Maxval uint64
	Depth int
	HeaderSize, FileSize uint64
}

// String will return the imageHeader as specified in the PNM specification.
	func (ih imageHeader) String() (s string) {
	if ih.Magic == "P1" || ih.Magic == "P4" {
		s = fmt.Sprintf("%s %d %d\n", ih.Magic, ih.Width, ih.Height)
	} else {
		s = fmt.Sprintf("%s %d %d %d\n",
			ih.Magic, ih.Width, ih.Height, ih.Maxval)
	}
	return
}

// Debug will print the values of the imageHeader, formatted for
// debugging output (multiline).
func (ih imageHeader) Debug() string {
	s := fmt.Sprintf("%-18s{magic:%s width:%d height:%d "+
		"bpp:%d headerlen:%d}\n", "image header:",
		ih.Magic, ih.Width, ih.Height, ih.Depth, ih.HeaderSize)
	s += fmt.Sprintf("%-18s%d\n", "bytes expected:",
		ih.Width*ih.Height*uint64(ih.Depth/8))
	s += fmt.Sprintf("%-18s%d - %d = %d", "bytes counted:",
		ih.FileSize, ih.HeaderSize, ih.FileSize-ih.HeaderSize)
	return s
}

func (ih imageHeader) calcRow() uint64 {
	return uint64(8*float32(ih.FileSize-ih.HeaderSize)/
		float32(uint64(ih.Depth)*ih.Width)+0.5)
}

// readPage reads a single page of the given io.Reader, sized 4K. If
// the page is less than 4K in size, the resulting slice will be
// truncated to the number of bytes available. If the page cannot be
// read (eg due to a previous error), page will hold the nil value.
func readPage(reader io.Reader) (page []byte, err error) {
	page = make([]byte, 4*1024)
	var n int
	n, err = reader.Read(page)
	if n > 0 {
		page = page[0:n]
	} else {
		page = nil
	}
	return
}

// parseHeader reads the given byte slice and tries to extract a PNM
// image header from it: it will read the magic number, the width,
// height, and (depending on the magic number) a maximum value, all
// separated by whitespaces (blanks, tabs, cr, and lf). Comments are
// handled as well acoording to the PNM specification (terminated by
// cr).
func parseHeader(page []byte) (ih imageHeader) {
	s:=2
	ih.Magic = string(page[0:s])
	field := Magic

	for s<len(page) && (field < Height ||
		(ih.Magic != "P1" && ih.Magic != "P4" && field<Maxval)) {
		// skip all whitespace
		c := page[s]
		if c==' ' || c=='\t' || c=='\r' || c=='\n' {
			s++
			continue
		}

		// skip comment lines
		if c == '#' {
			t := s
			for page[t]!='\r' && page[t]!='\n' && t<len(page) {
				t++
			}
			s = t
			continue
		}

		// read identifier
		t := s
		c = page[t]
		for c!=' ' && c!='\t' && c!='\r' && c!='\n' && t<len(page) {
			t++
			if t<len(page) {
				c = page[t]
			}
		}
		val, _ := strconv.ParseUint(string(page[s:t]), 10, 64)
		field++
		switch field {
		case Width:  ih.Width  = val
		case Height: ih.Height = val
		case Maxval: ih.Maxval = val
		}
		s = t
	}

	// set depth value
	if ih.Maxval<256 {
		ih.Depth=1
	} else {
		ih.Depth=2
	}
	switch ih.Magic {
	case "P2", "P5": ih.Depth*=8
	case "P3", "P6": ih.Depth*=24
	}

	ih.HeaderSize = uint64(s+1)
	return
}


// isConsole returns true if the given file descriptor is a console,
// false if it is a pipe
func isConsole(file *os.File) bool {
        var termios syscall.Termios
        _, _, err := syscall.Syscall6(syscall.SYS_IOCTL, file.Fd(),
		uintptr(syscall.TCGETS), uintptr(unsafe.Pointer(&termios)),
		0, 0, 0)
        return err == 0
}

func main() {
	var pnm [][]byte
	var filesize uint64 = 0
	stdin, stderr := os.Stdin, os.Stderr
	verbose := flag.Bool("verbose", false, "print image header info")
	flag.Parse()

	// read input in blocks of 4K
	page, _ := readPage(stdin)
	for page != nil {
		filesize += uint64(len(page))
		pnm = append(pnm, page)
		page, _ = readPage(stdin)
	}

	// parse image header
	var ih imageHeader
	if len(pnm) > 0 {
		ih = parseHeader(pnm[0])
		ih.FileSize = filesize
		if *verbose {
			fmt.Fprintln(stderr, ih.Debug())
		}
	} else {
		fmt.Fprintln(stderr, "Cannot read image data from stdin")
		os.Exit(1)
	}

	// evaluate image header
	relay := ih.Magic != "P4" && ih.Magic != "P5" && ih.Magic != "P6"
	if relay {
		fmt.Fprintf(stderr, "No support for format %s, "+
			"relaying image data\n", ih.Magic)
	}

	// determine whether PNM needs to be fixed
	if newHeight := ih.calcRow(); !relay && newHeight != ih.Height {
		fmt.Fprintf(stderr, "Fixing number of rows "+
			"from %d to %d\n", ih.Height, newHeight)
		ih.Height = newHeight
		pnm[0] = append([]byte(fmt.Sprint(ih)), pnm[0][ih.HeaderSize:]...)
	}

	// write image data to stdout
	if isConsole(os.Stdout) {
		fmt.Fprintln(stderr,
			"Not writing image data to console, exiting.")
	} else {
		for i := range pnm {
			fmt.Print(string(pnm[i]))
		}
	}
}
